# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
import urllib
import warnings
from argparse import Namespace
from pathlib import Path

import torch

import fm


def _has_regression_weights(model_name):
    """Return whether we expect / require regression weights;
    Right now that is all models except ESM-1v, ESM-IF, and partially trained ESM2 models"""
    return not ("esm1v" in model_name or "esm_if" in model_name or "270K" in model_name or "500K" in model_name or
                "checkpoint_best" in model_name or "RNA-FM" in model_name or "CDS-FM" in model_name)


# def load_model_and_alphabet(model_name):
#     if model_name.endswith(".pt"):  # treat as filepath
#         return load_model_and_alphabet_local(model_name)
#     else:
#         return load_model_and_alphabet_hub(model_name)


def load_hub_workaround(url, download_name=None):
    try:
        data = torch.hub.load_state_dict_from_url(url, progress=True, map_location='cpu', file_name=download_name)
    except RuntimeError:
        # Pytorch version issue - see https://github.com/pytorch/pytorch/issues/43106
        if download_name == None:
            fn = Path(url).name
        else:
            fn = download_name
        data = torch.load(
            f"{torch.hub.get_dir()}/checkpoints/{fn}",
            map_location="cpu",
        )
    return data


def load_regression_hub(model_name):
    url = f"https://dl.fbaipublicfiles.com/fair-esm/regression/{model_name}-contact-regression.pt"
    regression_data = load_hub_workaround(url)
    return regression_data


def _download_model_and_regression_data(model_name):
    url = f"https://dl.fbaipublicfiles.com/fair-esm/models/{model_name}.pt"
    model_data = load_hub_workaround(url)
    if _has_regression_weights(model_name):
        regression_data = load_regression_hub(model_name)
    else:
        regression_data = None
    return model_data, regression_data


# def load_model_and_alphabet_hub(model_name):
#     model_data, regression_data = _download_model_and_regression_data(model_name)
#     return load_model_and_alphabet_core(model_name, model_data, regression_data)


def load_model_and_alphabet_local(model_location, theme="protein"):
    """Load from local path. The regression weights need to be co-located"""
    model_location = Path(model_location)
    model_data = torch.load(str(model_location), map_location="cpu")
    model_name = model_location.stem
    if _has_regression_weights(model_name):
        regression_location = str(model_location.with_suffix("")) + "-contact-regression.pt"
        regression_data = torch.load(regression_location, map_location="cpu")
    else:
        regression_data = None
    return load_model_and_alphabet_core(model_name, model_data, regression_data, theme)


def has_emb_layer_norm_before(model_state):
    """Determine whether layer norm needs to be applied before the encoder"""
    return any(k.startswith("emb_layer_norm_before") for k, param in model_state.items())


def _load_model_and_alphabet_core_v1(model_data, theme="protein"):
    alphabet = fm.Alphabet.from_architecture(model_data["args"].arch, theme)

    if model_data["args"].arch == "roberta_large":
        # upgrade state dict
        pra = lambda s: "".join(s.split("encoder_")[1:] if "encoder" in s else s)
        prs1 = lambda s: "".join(s.split("encoder.")[1:] if "encoder" in s else s)
        prs2 = lambda s: "".join(
            s.split("sentence_encoder.")[1:] if "sentence_encoder" in s else s
        )
        model_args = {pra(arg[0]): arg[1] for arg in vars(model_data["args"]).items()}
        model_state = {prs1(prs2(arg[0])): arg[1] for arg in model_data["model"].items()}
        model_state["embed_tokens.weight"][alphabet.mask_idx].zero_()  # For token drop
        model_args["emb_layer_norm_before"] = has_emb_layer_norm_before(model_state)
        model_type = fm.BioBertModel   #ProteinBertModel

    else:
        raise ValueError("Unknown architecture selected")

    model = model_type(
        Namespace(**model_args),
        alphabet,
    )

    return model, alphabet, model_state


def _load_model_and_alphabet_core_v2(model_data):
    pass


def load_model_and_alphabet_core(model_name, model_data, regression_data=None, theme="protein"):
    if regression_data is not None:
        model_data["model"].update(regression_data["model"])

    if model_name.startswith("esm2"):
        model, alphabet, model_state = _load_model_and_alphabet_core_v2(model_data)
    else:
        model, alphabet, model_state = _load_model_and_alphabet_core_v1(model_data, theme)

    expected_keys = set(model.state_dict().keys())
    found_keys = set(model_state.keys())

    if regression_data is None:
        expected_missing = {"contact_head.regression.weight", "contact_head.regression.bias"}
        error_msgs = []
        missing = (expected_keys - found_keys) - expected_missing
        if missing:
            error_msgs.append(f"Missing key(s) in state_dict: {missing}.")
        unexpected = found_keys - expected_keys
        if unexpected:
            error_msgs.append(f"Unexpected key(s) in state_dict: {unexpected}.")

        if error_msgs:
            raise RuntimeError(
                "Error(s) in loading state_dict for {}:\n\t{}".format(
                    model.__class__.__name__, "\n\t".join(error_msgs)
                )
            )
        if expected_missing - found_keys:
            warnings.warn(
                "Regression weights not found, predicting contacts will not produce correct results."
            )

    model.load_state_dict(model_state, strict=regression_data is not None)

    return model, alphabet


def load_fm_model_and_alphabet_hub(model_name, theme="rna"):
    if model_name == "rna_fm_t12":
        url = f"https://proj.cse.cuhk.edu.hk/rnafm/api/download?filename=RNA-FM_pretrained.pth"
        model_data = load_hub_workaround(url, download_name="RNA-FM_pretrained.pth")
        #url = f"https://proj.cse.cuhk.edu.hk/rnafm/api/download?filename=RNA-FM_SS-ResNet.pth"
        #model_data = load_hub_workaround(url, download_name="RNA-FM_SS-ResNet.pth")
        regression_data = None
        return load_model_and_alphabet_core("rna-fm", model_data, regression_data, theme)
    elif model_name == "mrna_fm_t12":
        url = f"https://proj.cse.cuhk.edu.hk/rnafm/api/download?filename=file_mRNA-FM_pretrained.pth"
        model_data = load_hub_workaround(url, download_name="file_mRNA-FM_pretrained.pth")
        # url = f"https://proj.cse.cuhk.edu.hk/rnafm/api/download?filename=RNA-FM_SS-ResNet.pth"
        # model_data = load_hub_workaround(url, download_name="RNA-FM_SS-ResNet.pth")
        regression_data = None
        return load_model_and_alphabet_core("mrna-fm", model_data, regression_data, theme)
    else:
        raise Exception("Unknown model name: {}".format(model_name))


import os
def rna_fm_t12(model_location=None):
    if model_location is not None and os.path.exists(model_location):
        # local
        return load_model_and_alphabet_local(model_location, theme="rna")  # "./pretrained/RNA-FM_pretrained.pth"
    else:
        return load_fm_model_and_alphabet_hub("rna_fm_t12", theme="rna")


def mrna_fm_t12(model_location=None):
    if model_location is not None:
        return load_model_and_alphabet_local(model_location, theme="rna-3mer")
    else:
        return load_fm_model_and_alphabet_hub("mrna_fm_t12", theme="rna-3mer")




# encoding: utf-8
"""
@author:  Jiayang Chen
@contact: yjcmydkzgj@gmail.com
"""
import os
import torch
from .baseline import Baseline
from ..pretrained import load_hub_workaround

def build_model(cfg):
    model = Baseline(
        backbone_name=cfg.MODEL.BACKBONE_NAME,
        pairwise_predictor_name=cfg.MODEL.PAIRWISE_PREDICTOR_NAME,
        backbone_frozen=cfg.MODEL.BACKBONE_FROZEN,
        backbone_random_initialization=cfg.MODEL.BACKBONE_RANDOM_INITIALIZATION,
    )
    return model



def build_rnafm_resnet(type="ss", model_location=None):
    """
    :param type: for specific task type, like secondary structure prediction
    :return:
    """
    if type == "ss":
        model = Baseline(
            backbone_name="rna-fm",
            pairwise_predictor_name="pc-resnet_1_sym_first:r-ss",
            backbone_frozen=1
        )
        url = f"https://proj.cse.cuhk.edu.hk/rnafm/api/download?filename=RNA-FM-ResNet_PDB-All.pth"
    else:
        raise Exception("Unknown Model Type!")

    if model_location is None:
        model_state_dict = load_hub_workaround(url, download_name="RNA-FM-ResNet_PDB-All.pth")
        model.load_state_dict(model_state_dict)
    elif model_location is not None and os.path.exists(model_location):
        model_state_dict = torch.load(model_location)
        model.load_state_dict(model_state_dict)
    else:
        raise Exception("Wrong Local Location of Model Given")

    return model, model.backbone_alphabet







# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from .modules import (
    TransformerLayer,
    AxialTransformerLayer,
    LearnedPositionalEmbedding,
    SinusoidalPositionalEmbedding,
    RobertaLMHead,
    ESM1bLayerNorm,
    ContactPredictionHead,
)

from .axial_attention import RowSelfAttention, ColumnSelfAttention

# CJY copy from fairseq\modules\transformer_sentence_encoder.py for initialization
def init_bert_params(module):
    """
    Initialize the weights specific to the BERT Model.
    This overrides the default initializations depending on the specified arguments.
        1. If normal_init_linear_weights is set then weights of linear
           layer will be initialized using the normal distribution and
           bais will be set to the specified value.
        2. If normal_init_embed_weights is set then weights of embedding
           layer will be initialized using the normal distribution.
        3. If normal_init_proj_weights is set then weights of
           in_project_weight for MultiHeadAttention initialized using
           the normal distribution (to be validated).
    apply: (no effect)
    self.apply(init_bert_params)
    """

    def normal_(data):
        # with FSDP, module params will be on CUDA, so we cast them back to CPU
        # so that the RNG is consistent with and without FSDP
        data.copy_(
            data.cpu().normal_(mean=0.0, std=0.02).to(data.device)
        )

    if isinstance(module, nn.Linear):
        normal_(module.weight.data)
        if module.bias is not None:
            module.bias.data.zero_()
    if isinstance(module, nn.Embedding):
        normal_(module.weight.data)
        if module.padding_idx is not None:
            module.weight.data[module.padding_idx].zero_()
    from .multihead_attention import MultiheadAttention
    if isinstance(module, MultiheadAttention) or isinstance(module, RowSelfAttention) or isinstance(module,ColumnSelfAttention):
        normal_(module.q_proj.weight.data)
        normal_(module.k_proj.weight.data)
        normal_(module.v_proj.weight.data)

class RNABertModel(nn.Module):
    # referenced from ProteinBertModel (fairseq-esm)
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "--num_layers", default=36, type=int, metavar="N", help="number of layers"
        )
        parser.add_argument(
            "--embed_dim", default=1280, type=int, metavar="N", help="embedding dimension"
        )
        parser.add_argument(
            "--logit_bias", action="store_true", help="whether to apply bias to logits"
        )
        parser.add_argument(
            "--ffn_embed_dim",
            default=5120,
            type=int,
            metavar="N",
            help="embedding dimension for FFN",
        )
        parser.add_argument(
            "--attention_heads",
            default=20,
            type=int,
            metavar="N",
            help="number of attention heads",
        )

    def __init__(self, args, alphabet):
        super().__init__()
        self.args = args
        self.alphabet_size = len(alphabet)
        self.padding_idx = alphabet.padding_idx
        self.mask_idx = alphabet.mask_idx
        self.cls_idx = alphabet.cls_idx
        self.eos_idx = alphabet.eos_idx
        self.prepend_bos = alphabet.prepend_bos
        self.append_eos = alphabet.append_eos
        if self.args.arch == 'roberta_large':
            self.model_version = 'ESM-1b'
            self._init_submodules_esm1b()
        else:
            self.model_version = 'ESM-1'
            self._init_submodules_esm1()

    def _init_submodules_common(self):
        self.embed_tokens = nn.Embedding(
            self.alphabet_size, self.args.embed_dim, padding_idx=self.padding_idx
        )
        self.layers = nn.ModuleList(
            [
                TransformerLayer(
                    self.args.embed_dim, self.args.ffn_embed_dim, self.args.attention_heads,
                    add_bias_kv=(self.model_version != 'ESM-1b'),
                    use_esm1b_layer_norm=(self.model_version == 'ESM-1b'),
                )
                for _ in range(self.args.layers)
            ]
        )

        self.contact_head = ContactPredictionHead(
            self.args.layers * self.args.attention_heads,
            self.prepend_bos,
            self.append_eos,
            eos_idx=self.eos_idx,
        )

    def _init_submodules_esm1b(self):
        self._init_submodules_common()
        self.embed_scale = 1
        self.embed_positions = LearnedPositionalEmbedding(self.args.max_positions, self.args.embed_dim, self.padding_idx)
        self.emb_layer_norm_before = ESM1bLayerNorm(self.args.embed_dim)
        self.emb_layer_norm_after = ESM1bLayerNorm(self.args.embed_dim)
        self.lm_head = RobertaLMHead(
            embed_dim=self.args.embed_dim,
            output_dim=self.alphabet_size,
            weight=self.embed_tokens.weight
        )

    def _init_submodules_esm1(self):
        self._init_submodules_common()
        self.embed_scale = math.sqrt(self.args.embed_dim)
        self.embed_positions = SinusoidalPositionalEmbedding(self.args.embed_dim, self.padding_idx)
        self.embed_out = nn.Parameter(
            torch.zeros((self.alphabet_size, self.args.embed_dim))
        )
        self.embed_out_bias = None
        if self.args.final_bias:
            self.embed_out_bias = nn.Parameter(torch.zeros(self.alphabet_size))

    # CJY at 2021.10.20 add masked_tokens for lm_head
    def forward(self, tokens, repr_layers=[], need_head_weights=False, return_contacts=False, masked_tokens=None):
        if return_contacts:
            need_head_weights = True

        assert tokens.ndim == 2
        padding_mask = tokens.eq(self.padding_idx) # B, T

        x = self.embed_scale * self.embed_tokens(tokens)

        if getattr(self.args, 'token_dropout', False):
            x.masked_fill_((tokens == self.mask_idx).unsqueeze(-1), 0.0)
            # x: B x T x C
            mask_ratio_train = 0.15 * 0.8
            src_lengths = (~padding_mask).sum(-1)
            mask_ratio_observed = (tokens == self.mask_idx).sum(-1).float() / src_lengths
            x = x * (1 - mask_ratio_train) / (1 - mask_ratio_observed)[:, None, None]

        x = x + self.embed_positions(tokens)

        if self.model_version == 'ESM-1b':
            x = self.emb_layer_norm_before(x)
            if padding_mask is not None:
                x = x * (1 - padding_mask.unsqueeze(-1).type_as(x))

        repr_layers = set(repr_layers)
        hidden_representations = {}
        if 0 in repr_layers:
            hidden_representations[0] = x

        if need_head_weights:
            attn_weights = []

        # (B, T, E) => (T, B, E)
        x = x.transpose(0, 1)

        if not padding_mask.any():
            padding_mask = None

        for layer_idx, layer in enumerate(self.layers):
            x, attn = layer(x, self_attn_padding_mask=padding_mask, need_head_weights=need_head_weights)
            if (layer_idx + 1) in repr_layers:
                hidden_representations[layer_idx + 1] = x.transpose(0, 1)
            if need_head_weights:
                # (H, B, T, T) => (B, H, T, T)
                attn_weights.append(attn.transpose(1, 0))

        if self.model_version == 'ESM-1b':
            x = self.emb_layer_norm_after(x)
            x = x.transpose(0, 1) # (T, B, E) => (B, T, E)

            # last hidden representation should have layer norm applied
            if (layer_idx + 1) in repr_layers:
                hidden_representations[layer_idx + 1] = x
            # CJY at 2021.10.20 add masked_tokens for lm_head
            x = self.lm_head(x, masked_tokens)
        else:
            x = F.linear(x, self.embed_out, bias=self.embed_out_bias)
            x = x.transpose(0, 1) # (T, B, E) => (B, T, E)

        result = {"logits": x, "representations": hidden_representations}
        if need_head_weights:
            # attentions: B x L x H x T x T
            attentions = torch.stack(attn_weights, 1)
            if self.model_version == "ESM-1":
                # ESM-1 models have an additional null-token for attention, which we remove
                attentions = attentions[..., :-1]
            if padding_mask is not None:
                attention_mask = (1 - padding_mask.type_as(attentions))
                attention_mask = attention_mask.unsqueeze(1) * attention_mask.unsqueeze(2)
                attentions = attentions * attention_mask[:, None, None, :, :]
            result["attentions"] = attentions
            if return_contacts:
                contacts = self.contact_head(tokens, attentions)
                result["contacts"] = contacts

        return result

    def predict_contacts(self, tokens):
        return self(tokens, return_contacts=True)["contacts"]

    @property
    def num_layers(self):
        return self.args.layers


class MSATransformer(nn.Module):

    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "--num_layers", default=12, type=int, metavar="N", help="number of layers"
        )
        parser.add_argument(
            "--embed_dim", default=768, type=int, metavar="N", help="embedding dimension"
        )
        parser.add_argument(
            "--logit_bias", action="store_true", help="whether to apply bias to logits"
        )
        parser.add_argument(
            "--ffn_embed_dim",
            default=3072,
            type=int,
            metavar="N",
            help="embedding dimension for FFN",
        )
        parser.add_argument(
            "--attention_heads",
            default=12,
            type=int,
            metavar="N",
            help="number of attention heads",
        )
        parser.add_argument(
            "--dropout",
            default=0.1,
            type=float,
            help="Dropout to apply."
        )
        parser.add_argument(
            "--attention_dropout",
            default=0.1,
            type=float,
            help="Dropout to apply."
        )
        parser.add_argument(
            "--activation_dropout",
            default=0.1,
            type=float,
            help="Dropout to apply."
        )
        parser.add_argument(
            "--max_tokens_per_msa",
            default=2 ** 14,
            type=int,
            help=(
                "Used during inference to batch attention computations in a single "
                "forward pass. This allows increased input sizes with less memory."
            ),
        )

    def __init__(self, args, alphabet):
        super().__init__()
        self.args = args
        self.alphabet_size = len(alphabet)
        self.padding_idx = alphabet.padding_idx
        self.mask_idx = alphabet.mask_idx
        self.cls_idx = alphabet.cls_idx
        self.eos_idx = alphabet.eos_idx
        self.prepend_bos = alphabet.prepend_bos
        self.append_eos = alphabet.append_eos

        self.embed_tokens = nn.Embedding(
            self.alphabet_size, self.args.embed_dim, padding_idx=self.padding_idx
        )

        if getattr(args, "embed_positions_msa", False):
            self.msa_position_embedding = nn.Parameter(
                0.01 * torch.randn(1, 1024, 1, 1),
                requires_grad=True,
            )
        else:
            self.register_parameter("msa_position_embedding", None)

        self.dropout_module = nn.Dropout(self.args.dropout)
        self.layers = nn.ModuleList(
            [
                AxialTransformerLayer(
                    self.args.embed_dim,
                    self.args.ffn_embed_dim,
                    self.args.attention_heads,
                    self.args.dropout,
                    self.args.attention_dropout,
                    self.args.activation_dropout,
                    getattr(self.args, "max_tokens_per_msa", self.args.max_tokens),
                )
                for _ in range(self.args.layers)
            ]
        )

        self.contact_head = ContactPredictionHead(
            self.args.layers * self.args.attention_heads,
            self.prepend_bos,
            self.append_eos,
            eos_idx=self.eos_idx,
        )
        self.embed_positions = LearnedPositionalEmbedding(
            self.args.max_positions, self.args.embed_dim, self.padding_idx,
        )
        self.emb_layer_norm_before = ESM1bLayerNorm(self.args.embed_dim)
        self.emb_layer_norm_after = ESM1bLayerNorm(self.args.embed_dim)
        self.lm_head = RobertaLMHead(
            embed_dim=self.args.embed_dim,
            output_dim=self.alphabet_size,
            weight=self.embed_tokens.weight
        )

    def forward(
        self,
        tokens,
        repr_layers=[],
        need_head_weights=False,
        return_contacts=False
    ):
        if return_contacts:
            need_head_weights = True

        assert tokens.ndim == 3
        batch_size, num_alignments, seqlen = tokens.size()
        padding_mask = tokens.eq(self.padding_idx)  # B, R, C
        if not padding_mask.any():
            padding_mask = None

        x = self.embed_tokens(tokens)
        x += self.embed_positions(
            tokens.view(batch_size * num_alignments, seqlen)
        ).view(x.size())
        if self.msa_position_embedding is not None:
            if x.size(1) > 1024:
                raise RuntimeError(
                    "Using model with MSA position embedding trained on maximum MSA "
                    f"depth of 1024, but received {x.size(1)} alignments."
                )
            x += self.msa_position_embedding[:, :num_alignments]

        x = self.emb_layer_norm_before(x)

        x = self.dropout_module(x)

        if padding_mask is not None:
            x = x * (1 - padding_mask.unsqueeze(-1).type_as(x))

        repr_layers = set(repr_layers)
        hidden_representations = {}
        if 0 in repr_layers:
            hidden_representations[0] = x

        if need_head_weights:
            row_attn_weights = []
            col_attn_weights = []

        # B x R x C x D -> R x C x B x D
        x = x.permute(1, 2, 0, 3)

        for layer_idx, layer in enumerate(self.layers):
            x = layer(
                x,
                self_attn_padding_mask=padding_mask,
                need_head_weights=need_head_weights,
            )
            if need_head_weights:
                x, col_attn, row_attn = x
                # H x C x B x R x R -> B x H x C x R x R
                col_attn_weights.append(col_attn.permute(2, 0, 1, 3, 4))
                # H x B x C x C -> B x H x C x C
                row_attn_weights.append(row_attn.permute(1, 0, 2, 3))
            if (layer_idx + 1) in repr_layers:
                hidden_representations[layer_idx + 1] = x.permute(2, 0, 1, 3)

        x = self.emb_layer_norm_after(x)
        x = x.permute(2, 0, 1, 3)  # R x C x B x D -> B x R x C x D

        # last hidden representation should have layer norm applied
        if (layer_idx + 1) in repr_layers:
            hidden_representations[layer_idx + 1] = x
        x = self.lm_head(x)

        result = {"logits": x, "representations": hidden_representations}
        if need_head_weights:
            # col_attentions: B x L x H x C x R x R
            col_attentions = torch.stack(col_attn_weights, 1)
            # row_attentions: B x L x H x C x C
            row_attentions = torch.stack(row_attn_weights, 1)
            result["col_attentions"] = col_attentions
            result["row_attentions"] = row_attentions
            if return_contacts:
                contacts = self.contact_head(
                    tokens, row_attentions
                )
                result["contacts"] = contacts

        return result

    def predict_contacts(self, tokens):
        return self(tokens, return_contacts=True)["contacts"]

    @property
    def num_layers(self):
        return self.args.layers

    def max_tokens_per_msa_(self, value: int) -> None:
        """ The MSA Transformer automatically batches attention computations when
        gradients are disabled to allow you to pass in larger MSAs at test time than
        you can fit in GPU memory. By default this occurs when more than 2^14 tokens
        are passed in the input MSA. You can set this value to infinity to disable
        this behavior.
        """
        for module in self.modules():
            if isinstance(module, (RowSelfAttention, ColumnSelfAttention)):
                module.max_tokens_per_msa = value



import torch
from torch import nn
from ..downstream_module import DownStreamModule

class LDotProduct(DownStreamModule):
    """
    contact predictor with attention
    """
    def __init__(self, backbone_args, backbone_alphabet, num_classes=1, class_embed_dim_out=512, symmetric=False, depth_reduction="mean"):
        """
        :param depth_reduction: mean, first, (adaptive)
        """
        super().__init__(backbone_args, backbone_alphabet, depth_reduction,
                         need_token=False, need_attention=[], need_embedding=[12])
        self.embed_dim_in = self.backbone_args.embed_dim
        self.attention_heads = self.backbone_args.attention_heads

        self.num_classes = num_classes
        self.symmetric = symmetric

        self.class_embed_dim_out = class_embed_dim_out

        self.q_proj = nn.Linear(self.embed_dim_in, self.num_classes * self.class_embed_dim_out)
        self.k_proj = nn.Linear(self.embed_dim_in, self.num_classes * self.class_embed_dim_out)

    def forward(self, tokens, inputs):
        embeddings = inputs["embedding"]
        # remove auxiliary tokens
        embeddings, padding_masks = self.remove_pend_tokens_1d(tokens, embeddings)

        if len(embeddings.size()) == 3:       # for seq
            batch_size, seqlen, hiddendim = embeddings.size()
        elif len(embeddings.size()) == 4:     # for msa
            batch_size, depth, seqlen, hiddendim = embeddings.size()
            embeddings = self.msa_depth_reduction(embeddings, padding_masks)
        else:
            raise Exception("Unknown Embedding Type!")

        # linear projection
        q = self.q_proj(embeddings).view(batch_size, seqlen, self.num_classes, self.class_embed_dim_out)
        k = self.k_proj(embeddings).view(batch_size, seqlen, self.num_classes, self.class_embed_dim_out)

        # short-cut
        #embeddings = embeddings.unsqueeze(-2).expand_as(q)
        #q = embeddings + q
        #k = embeddings + k

        output = torch.einsum("bick,bjck->bcij", q, k)

        if self.symmetric == True:
            # reverse and add
            #output = output + output.permute(0, 1, 3, 2)
            #output = torch.squeeze(output, dim=1)

            # max
            #output = torch.max(output, output.permute(0, 1, 3, 2))
            #output = torch.squeeze(output, dim=1)

            upper_triangular_output = torch.triu(output)
            lower_triangular_output = torch.triu(output, diagonal=1).permute(0, 1, 3, 2)
            output = upper_triangular_output + lower_triangular_output
            output = torch.squeeze(output, dim=1)

        return output

    @classmethod
    def create_module_with_name(cls, module_name, backbone_args, backbone_alphabet):
        _, num_class, symmetric, depth_reduction = module_name.split("_")
        num_class = int(num_class)
        if symmetric == "sym":
            symmetric = True
        elif symmetric == "asym":
            symmetric = False
        else:
            raise Exception("Wrong Symmetric Type!")
        module = cls(backbone_args, backbone_alphabet,
                     num_classes=num_class, symmetric=symmetric, depth_reduction=depth_reduction)
        return module



# encoding: utf-8
"""
@author:  Jiayang Chen
@contact: yjcmydkzgj@gmail.com
"""
from torch import nn
import torch
from functools import wraps




class DownStreamModule(nn.Module):
    """
    base contact predictor for msa
    """
    def __init__(self, backbone_args, backbone_alphabet, depth_reduction="none",
                 need_token=False, need_attention=[], need_embedding=[12], need_extrafeat=[]):
        super().__init__()
        self.backbone_args = backbone_args
        self.backbone_alphabet = backbone_alphabet

        self.prepend_bos = self.backbone_alphabet.prepend_bos
        self.append_eos = self.backbone_alphabet.append_eos
        self.bos_idx = self.backbone_alphabet.cls_idx
        self.eos_idx = self.backbone_alphabet.eos_idx
        if self.append_eos and self.eos_idx is None:
            raise ValueError("Using an alphabet with eos token, but no eos token was passed in.")
        self.pad_idx = self.backbone_alphabet.padding_idx

        self.embed_dim = self.backbone_args.embed_dim
        self.attention_heads = self.backbone_args.attention_heads

        self.depth_reduction = depth_reduction
        if self.depth_reduction == "attention":
            self.msa_embed_dim_in = self.embed_dim
            self.msa_embed_dim_out = self.embed_dim // self.attention_heads
            self.msa_q_proj = nn.Linear(self.msa_embed_dim_in, self.msa_embed_dim_out)
            self.msa_k_proj = nn.Linear(self.msa_embed_dim_in, self.msa_embed_dim_out)

        self.input_type = {
            "token": need_token,
            "attention": need_attention,
            "embedding": need_embedding,
            "extra-feat": need_extrafeat,
        }


    def remove_pend_tokens_1d(self, tokens, seqs):
        """
        :param tokens:
        :param seqs: must be shape of [B, ..., L, E]    # seq: [B, L, E]; msa: [B, D, L, E]
        :return:
        """
        padding_masks = tokens.ne(self.pad_idx)

        # remove eos token  （suffix first）
        if self.append_eos:     # default is right
            eos_masks = tokens.ne(self.eos_idx)
            eos_pad_masks = (eos_masks & padding_masks).to(seqs)
            seqs = seqs * eos_pad_masks.unsqueeze(-1)
            seqs = seqs[:, ..., :-1, :]
            padding_masks = padding_masks[:, ..., :-1]

        # remove bos token
        if self.prepend_bos:    # default is left
            seqs = seqs[:, ..., 1:, :]
            padding_masks = padding_masks[:, ..., 1:]

        if not padding_masks.any():
            padding_masks = None

        return seqs, padding_masks

    def remove_pend_tokens_2d(self, tokens, maps):
        """
        :param tokens:
        :param maps: shape of [B, L, L, ...]
        :return:
        """
        padding_masks = tokens.ne(self.pad_idx)

        # remove eos token  （suffix first）
        if self.append_eos:  # default is right
            eos_masks = tokens.ne(self.eos_idx)
            eos_pad_masks = (eos_masks & padding_masks).to(maps)
            eos_pad_masks = eos_pad_masks.unsqueeze(1) * eos_pad_masks.unsqueeze(2)
            maps = maps * eos_pad_masks.unsqueeze(-1)
            maps = maps[:, :-1, :-1, ...]
            padding_masks = padding_masks[:, :-1, ...]

        # remove bos token
        if self.prepend_bos:  # default is left
            maps = maps[:, 1:, 1:, ...]
            padding_masks = padding_masks[:, 1:, ...]

        if not padding_masks.any():
            padding_masks = None

        return maps, padding_masks

    def msa_depth_reduction(self, embeddings, padding_masks):
        """
        :param embeddings:  B,
        :param padding_masks:
        :return:
        """
        if self.depth_reduction == "first":
            embeddings = embeddings[:, 0, :, :]
        elif self.depth_reduction == "mean":
            embeddings = torch.mean(embeddings, dim=1)
        elif self.depth_reduction == "attention":
            msa_q = self.msa_q_proj(embeddings[:, 0, :, :])  # first query
            msa_k = self.msa_k_proj(embeddings)  # all keys
            if padding_masks is not None:
                # Zero out any padded aligned positions - this is important since
                # we take a sum across the alignment axis.
                msa_q = msa_q * (1 - padding_masks[:, 0, :].unsqueeze(-1).type_as(msa_q))
            depth_attn_weights = torch.einsum("bld,bjld->bj", msa_q, msa_k)
            depth_attn_weights = torch.softmax(depth_attn_weights, dim=1)
            embeddings = torch.sum(embeddings * depth_attn_weights.unsqueeze(-1).unsqueeze(-1), dim=1)
        else:
            raise Exception("Wrong Depth Reduction Type")

        return embeddings


# encoding: utf-8
"""
@author:  Jiayang Chen
@contact: yjcmydkzgj@gmail.com
"""

import torch

from .backbones import choose_backbone
from .pairwise_predictor import choose_pairwise_predictor

from .weights_init import *
from ptflops import get_model_complexity_info
from torch.cuda.amp import autocast

class Baseline(nn.Module):
    def __init__(self,
                 backbone_name,
                 seqwise_predictor_name="none",
                 elewise_predictor_name="none",
                 pairwise_predictor_name="none",
                 backbone_frozen=0,
                 backbone_random_initialization=False,   # because most of models used in this pipeline are pretrained
                 ):
        super(Baseline, self).__init__()
        # 0.Configuration
        self.backbone_name = backbone_name

        self.seqwise_predictor_name = seqwise_predictor_name
        self.elewise_predictor_name = elewise_predictor_name
        self.pairwise_predictor_name = pairwise_predictor_name

        self.backbone_frozen = backbone_frozen
        self.backbone_frozen_output_cache = {}

        self.backbone_random_initialization = backbone_random_initialization

        # just for reference
        self.target_info = {
            "p-ss": {"num_class": 3, "symmetric": None},        # or 8
            "p-contact": {"num_class": 2, "symmetric": True},
            "p-dist-bin": {"num_class": 37, "symmetric": True},
            "p-omega-bin": {"num_class": 25, "symmetric": True},
            "p-phi-bin": {"num_class": 13, "symmetric": False},
            "p-theta-bin": {"num_class": 25, "symmetric": False},
            "r-ss": {"num_class": 2, "symmetric": True},
            "ts-rgs": {"num_class": -1, "symmetric": None},
            "ts-cls": {"num_class": 2, "symmetric": None},
        }

        # 1.Build backbone
        self.backbone, self.backbone_alphabet = choose_backbone(self.backbone_name)
        self.backbone.lm_head.requires_grad_(False)

        self.backbone_control_info = {"return_contacts": False, "need_head_weights": False, "repr_layers": []}

        # 2.Build downstream modules
        self.downstream_modules = nn.ModuleDict()

        # the conposition of each name are "predictor_type" + ":" + "target_name"
        if self.pairwise_predictor_name != "none":
            for pp_name in self.pairwise_predictor_name.split(" "):
                predictor_type, target_name = pp_name.split(":")
                self.downstream_modules[pp_name] = choose_pairwise_predictor(
                    predictor_type, self.backbone.args, self.backbone_alphabet
                )
                self.update_backbone_contral_info(self.downstream_modules[pp_name].input_type)

        # to generate embeddings
        if self.seqwise_predictor_name == "none" and self.elewise_predictor_name == "none" and self.pairwise_predictor_name == "none":
            self.backbone_control_info["repr_layers"] = [12]

        # Initialization of parameters
        # omit backbone initialization
        if self.backbone_random_initialization == True:
            self.backbone.apply(weights_init_kaiming)
            #self.backbone.apply(weights_init_classifier)
        for key in self.downstream_modules.keys():
            self.downstream_modules[key].apply(weights_init_kaiming)
            self.downstream_modules[key].apply(weights_init_classifier)   # 最优 reduction这块还挺重要的

    #@autocast()
    def forward(self, data):
        x = data["token"]
        need_head_weights = self.backbone_control_info["need_head_weights"]
        repr_layers = self.backbone_control_info["repr_layers"]
        return_contacts = self.backbone_control_info["return_contacts"]

        if self.backbone_frozen != -1:
            if self.backbone_frozen == 1:
                self.backbone.eval()
                with torch.no_grad():
                    results = self.backbone(x, need_head_weights=need_head_weights,
                                            repr_layers=repr_layers, return_contacts=return_contacts)
            elif self.backbone_frozen == 2:
                need_forward = False
                for des in data["description"]:
                    if des not in self.backbone_frozen_output_cache:
                        need_forward = True

                if need_forward == True:
                    self.backbone.eval()
                    with torch.no_grad():
                        results = self.backbone(x, need_head_weights=need_head_weights,
                                                repr_layers=repr_layers, return_contacts=return_contacts)
                    self.save_backbone_output_cache(data, results)
                else:
                    results = self.load_backbone_output_cache(data)
            else:  # 0
                results = self.backbone(x, need_head_weights=need_head_weights,
                                        repr_layers=repr_layers, return_contacts=return_contacts)
        else:
            results = {}

        for key in self.downstream_modules.keys():
            key_info = key.split(":")
            predictor_type, target_name = key_info[0], key_info[1]
            ds_module_input = self.fetch_ds_module_input(data, results, self.downstream_modules[key].input_type)
            results[target_name] = self.downstream_modules[key](x, ds_module_input).float()    # .float() for fp16 to fp32

        return results


    def update_backbone_contral_info(self, ds_module_input_type):
        """
        :param ds_module_input_type: the downstream modules may need different output from backbone. This is a dict of
        control parameters for backbone.
        :return: update self.backbone_control_info
        """
        for key in ds_module_input_type.keys():
            if key == "token":
                pass
            elif key == "attention":
                if ds_module_input_type[key] != []:
                    self.backbone_control_info["need_head_weights"] = True
            elif key == "embedding":
                if isinstance(ds_module_input_type[key], int):
                    layer_i = ds_module_input_type[key]
                    if layer_i not in self.backbone_control_info["repr_layers"]:
                        self.backbone_control_info["repr_layers"].append(ds_module_input_type[key])
                elif isinstance(ds_module_input_type[key], list) and len(ds_module_input_type[key]) == 1:
                    layer_i = ds_module_input_type[key][0]
                    if layer_i not in self.backbone_control_info["repr_layers"]:
                        self.backbone_control_info["repr_layers"].append(layer_i)
                else:
                    for layer_i in ds_module_input_type[key]:
                        if layer_i not in self.backbone_control_info["repr_layers"]:
                            self.backbone_control_info["repr_layers"].append(ds_module_input_type[key])
            elif key == "extra-feat":  # features from third parties
                pass
            else:
                raise Exception("Unknown Keys for DS Module Input")


    def fetch_ds_module_input(self, data, backbone_results, ds_module_input_type):
        """
        The input of a specific downstream module may be a subset of backbone's output and we should pick them up into a
        sub_results dictionary and pass it into the downstream stask.
        :param tokens: output of this ds_input, there must be tokens as the first parameter for generate mask.
        :param backbone_results:
        :param ds_module_input_type:
        :return:
        """
        tokens = data["token"]
        ds_input = {}
        for key in ds_module_input_type.keys():
            if key == "token":
                if ds_module_input_type[key] == True:
                    ds_input[key] = tokens
            elif key == "attention":
                if ds_module_input_type[key] != []:
                    attention = backbone_results["attentions"] if "attentions" in backbone_results else backbone_results["row_attentions"]
                    ds_input[key] = attention
            elif key == "embedding":
                if isinstance(ds_module_input_type[key], int):
                    ds_input[key] = backbone_results["representations"][ds_module_input_type[key]]
                elif isinstance(ds_module_input_type[key], list) and len(ds_module_input_type[key]) == 1:
                    ds_input[key] = backbone_results["representations"][ds_module_input_type[key][0]]
                else:
                    for layer_i in ds_module_input_type[key]:
                        ds_input[key] = backbone_results["representations"][layer_i]
            elif key == "extra-feat":  # features from third parties
                for k in ds_module_input_type[key]:
                    ds_input[k] = data[k]
            else:
                raise Exception("Unknown Keys for DS Module Input")
        return ds_input

    def save_backbone_output_cache(self, data, results):
        name_list = data["description"]
        # save cache
        for i, descritption in enumerate(name_list):
            temp_results = {}
            for key in results.keys():
                if isinstance(results[key], dict):
                    temp_results[key] = {}
                    for sub_key in results[key].keys():
                        if key == "representations" and len(results[key][sub_key][i].shape) == 2:
                            # we should save tensor without padding (reserve bos & eos)
                            lrange = (0, self.backbone_alphabet.prepend_bos + data["length"][i] + self.backbone_alphabet.append_eos)
                            temp_results[key][sub_key] = results[key][sub_key][i][lrange[0]:lrange[1]].detach().cpu()
                else:
                    temp_results[key] = results[key][i].detach().cpu()
                    # attention need to be implemented
            self.backbone_frozen_output_cache[descritption] = temp_results

    def load_backbone_output_cache(self, data):
        name_list = data["description"]
        # load cache
        results = {}
        for key in self.backbone_frozen_output_cache[name_list[0]].keys():
            if isinstance(self.backbone_frozen_output_cache[name_list[0]][key], dict):
                results[key] = {}
                for sub_key in self.backbone_frozen_output_cache[name_list[0]][key].keys():
                    temp_results = []
                    for i, descritption in enumerate(name_list):
                        temp_results.append(self.backbone_frozen_output_cache[descritption][key][sub_key].to(data["token"].float()))
                    try:
                        results[key][sub_key] = torch.stack(temp_results, dim=0)
                    except:
                        results[key][sub_key] = self.stack_variable_length_tensors(temp_results)
            else:
                temp_results = []
                for i, descritption in enumerate(name_list):
                    temp_results.append(self.backbone_frozen_output_cache[descritption][key].to(data["token"].float()))
                try:
                    results[key] = torch.stack(temp_results, dim=0)
                except:
                    results[key] = self.stack_variable_length_tensors(temp_results)
        return results

    def stack_variable_length_tensors(self, vl_tensors):
        dim_value = []
        for t in vl_tensors:
            dim_value.append(t.shape)
        dim_max_value = torch.Tensor(dim_value).max(dim=0)[0].int().numpy().tolist()

        pad_tensors = []
        for t in vl_tensors:
            pad_shape = []
            for d_i, d_max in enumerate(dim_max_value):
                pad_shape.append(d_max-t.shape[d_i])
                pad_shape.append(0)
            pad_shape.reverse()   # order need from left to top
            pad_tensors.append(torch.nn.functional.pad(t, pad_shape, mode="constant"))
        output = torch.stack(pad_tensors, dim=0)
        return output

    # load parameter
    def load_param(self, load_choice, model_path):
        param_dict = torch.load(model_path, map_location="cpu")  #["model"]
        if param_dict.get("model") is not None and param_dict.get("args") is not None:
            print("Does not reload weights from official pre-trained file!")
            return 1

        if load_choice == "backbone":
            base_dict = self.backbone.state_dict()
            for i in param_dict:
                module_name = i.replace("backbone.", "")
                if module_name not in self.backbone.state_dict():
                    print("Cannot load %s, Maybe you are using incorrect framework"%i)
                    continue
                self.backbone.state_dict()[module_name].copy_(param_dict[i])
            print("Complete Load Weight")

        elif load_choice == "overall":
            map_dict = {}
            overall_dict = self.state_dict()
            for key in overall_dict.keys():
                if ":" in key:
                    sub_keys = key.split(":")
                    module_master_name = sub_keys[0]
                    module_branch_name = sub_keys[1].split(".", 1)[1]
                    module_name = module_master_name + "." + module_branch_name
                else:
                    module_name = key
                map_dict[module_name] = {"tar": key}

            for key in param_dict.keys():
                if ":" in key:
                    sub_keys = key.split(":")
                    module_master_name = sub_keys[0]
                    module_branch_name = sub_keys[1].split(".", 1)[1]
                    module_name = module_master_name + "." + module_branch_name
                else:
                    module_name = key
                try:
                    map_dict[module_name]["src"] = key
                except:
                    print("Cannot load %s, Maybe you are using incorrect framework" % key)

            for i in map_dict.keys():
                self.state_dict()[map_dict[i]["tar"]].copy_(param_dict[map_dict[i]["src"]])

            """
            for i in param_dict:
                if i in self.state_dict():
                    self.state_dict()[i].copy_(param_dict[i])
                elif "base."+i in self.state_dict():
                    self.state_dict()["base."+i].copy_(param_dict[i])
                elif "backbone."+i in self.state_dict():
                    self.state_dict()["backbone."+i].copy_(param_dict[i])
                elif i.replace("base", "backbone") in self.state_dict():
                    self.state_dict()[i.replace("base", "backbone")].copy_(param_dict[i])
                else:
                    print("Cannot load %s, Maybe you are using incorrect framework"%i)
                    continue
            """
            print("Complete Load Weight")

        elif load_choice == "none":
            print("Do not reload Weight by myself.")


    def count_param(model, input_shape=(3, 224, 224)):
        with torch.cuda.device(0):
            flops, params = get_model_complexity_info(model, input_shape, as_strings=True, print_per_layer_stat=True)
            print('{:<30}  {:<8}'.format('Computational complexity: ', flops))
            print('{:<30}  {:<8}'.format('Number of parameters: ', params))
            return ('{:<30}  {:<8}'.format('Computational complexity: ', flops)) + (
                '{:<30}  {:<8}'.format('Number of parameters: ', params))


# encoding: utf-8
"""
@author:  Jiayang Chen
@contact: yjcmydkzgj@gmail.com
"""

from .pairwise_concat import PairwiseConcat, PairwiseConcatWithResNet
from .linear_dot_product import LDotProduct

def choose_pairwise_predictor(pairwise_predictor_name, backbone_args, backbone_alphabet):
    """
    :param pairwise_predictor_name:
    :param backbone_args:
    :param backbone_alphabet:
    :return:
    Methods Notes:
    1.Tied Row Attention - Based
    "build-in": the logistic regression based on multi-layer tied row attention (build-in)
    "LR": the logistic regression based on multi-layer tied row attention (external)
    "MultiLayerTiedRowAttention-$lr1,$lr2,...": could select layers and set num_classes additionally
                                                $lr in [1,12]
    2.Embedding - Based
    $reduction in {"first", "mean", "attention"}
    'InnerProduct-$reduction': PairwiseInnerProduct
    'PairwiseConcat-$reduction': PairwiseConcat + Linear
    'SelfAttention-$reduction': SelfAttention
    """
    if pairwise_predictor_name == 'build-in':
        pairwise_predictor = None
        #backbone_control_parameter_dict = {"return_contacts": True, "need_head_weights": True}
    elif 'pairwise-concat' in pairwise_predictor_name:
        pairwise_predictor = PairwiseConcat.create_module_with_name(
            pairwise_predictor_name, backbone_args, backbone_alphabet
        )
    elif 'pc-resnet' in pairwise_predictor_name:
        pairwise_predictor = PairwiseConcatWithResNet.create_module_with_name(
            pairwise_predictor_name, backbone_args, backbone_alphabet
        )
    elif 'linear-dot-product' in pairwise_predictor_name:
        pairwise_predictor = LDotProduct.create_module_with_name(
            pairwise_predictor_name, backbone_args, backbone_alphabet
        )
    elif pairwise_predictor_name == "none":
        pairwise_predictor = None
        print("Without Independent Contact Predictor!")
    else:
        raise Exception("Wrong Backbone Type!")

    return pairwise_predictor


import torch
from torch import nn
from ..downstream_module import DownStreamModule

from collections import OrderedDict


class Lin2D(nn.Module):
    def __init__(self, in_feat, out_feat):
        super().__init__()
        self.linear = nn.Linear(in_feat, 256)
        self.relu = nn.ReLU()
        self.linear2 = nn.Linear(256, out_feat)

    def forward(self, inputs):
        inputs = inputs.permute(0, 2, 3, 1)

        outputs = self.linear2(self.relu(self.linear(inputs)))
        #outputs = self.linear(inputs)

        outputs = outputs.permute(0, 3, 1, 2)

        return outputs


class PairwiseConcat(DownStreamModule):
    """
    contact predictor with pairwise concat
    """
    def __init__(self, backbone_args, backbone_alphabet, num_classes,
                 symmetric=False, embed_reduction=-1, depth_reduction="mean"):
        """
        :param depth_reduction: mean, first, (adaptive)
        """
        super().__init__(backbone_args, backbone_alphabet, depth_reduction,
                         need_token=False, need_attention=[], need_embedding=[12], need_extrafeat=[])
        self.embed_dim_in = self.backbone_args.embed_dim
        self.attention_heads = self.backbone_args.attention_heads

        self.num_classes = num_classes
        self.symmetric = symmetric

        self.embed_reduction = embed_reduction
        if self.embed_reduction != -1:
            self.embed_dim_out = self.embed_reduction
            self.pre_reduction = nn.Linear(self.embed_dim_in, self.embed_dim_out)
        else:
            self.embed_dim_out = self.embed_dim_in
            self.pre_reduction = None

        self.proj = Lin2D(self.embed_dim_out * 2, self.num_classes)  #nn.Conv2d(self.embed_dim_out * 2, self.num_classes, kernel_size=1)

    def forward(self, tokens, inputs):
        embeddings = inputs["embedding"]
        # remove auxiliary tokens
        embeddings, padding_masks = self.remove_pend_tokens_1d(tokens, embeddings)

        if len(embeddings.size()) == 3:       # for seq
            batch_size, seqlen, hiddendim = embeddings.size()                    # B, L, E
        elif len(embeddings.size()) == 4:     # for msa
            batch_size, depth, seqlen, hiddendim = embeddings.size()
            embeddings = self.msa_depth_reduction(embeddings, padding_masks)     # B, L, E
        else:
            raise Exception("Unknown Embedding Type!")

        # embedding dim reduction
        if self.embed_reduction != -1:
            embeddings = self.pre_reduction(embeddings)
            hiddendim = self.embed_reduction

        # inner product
        # embedding_T = embedding.permute(0, 2, 1)
        # inner_product_map = torch.matmul(embedding, embedding_T)
        # contact_map = self.activation(inner_product_map)
        # with BN
        # inner_product_map = torch.matmul(embedding, embedding_T).unsqueeze(1)
        # inner_product_map = self.activation(self.BN(inner_product_map)).squeeze(1)

        # cosine similarity
        embeddings = embeddings.unsqueeze(2).expand(batch_size, seqlen, seqlen, hiddendim)
        embedding_T = embeddings.permute(0, 2, 1, 3)
        pairwise_concat_embedding = torch.cat([embeddings, embedding_T], dim=3)
        pairwise_concat_embedding = pairwise_concat_embedding.permute(0, 3, 1, 2)

        output = self.proj(pairwise_concat_embedding)

        if self.symmetric == True:
            # reverse and add
            #output = output + output.permute(0, 1, 3, 2)
            #output = torch.squeeze(output, dim=1)

            # max
            #output = torch.max(output, output.permute(0, 1, 3, 2))
            #output = torch.squeeze(output, dim=1)

            # half
            upper_triangular_output = torch.triu(output)
            lower_triangular_output = torch.triu(output, diagonal=1).permute(0, 1, 3, 2)
            output = upper_triangular_output + lower_triangular_output
            output = torch.squeeze(output, dim=1)

        return output

    @classmethod
    def create_module_with_name(cls, module_name, backbone_args, backbone_alphabet):
        _, num_class, symmetric, depth_reduction = module_name.split("_")
        num_class = int(num_class)
        if symmetric == "sym":
            symmetric = True
        elif symmetric == "asym":
            symmetric = False
        else:
            raise Exception("Wrong Symmetric Type!")
        module = cls(backbone_args, backbone_alphabet,
                     num_classes=num_class, symmetric=symmetric, depth_reduction=depth_reduction)
        return module


class PairwiseConcatWithResNet(PairwiseConcat):
    """
    contact predictor with pairwise concat + resnet
    reproduce from msa tranformer
    """
    def __init__(self, backbone_args, backbone_alphabet, num_classes,
                 symmetric=False, embed_reduction=128, depth_reduction="mean"):
        """
        :param depth_reduction: mean, first, (adaptive)
        """
        super().__init__(backbone_args, backbone_alphabet, num_classes, symmetric, embed_reduction, depth_reduction,)

        # num parameter 640 * 128 + 256 * n + 32 * 9 * n * n + 9 * n = 81920 + 265n + 288n^2
        # n = 32:  81920 + 8480 + 294912 = 385,312
        # n = 64:  81920 + 16960 + 1179648 = 1,278,528
        # n = 96:  81920 + 25440 + 2654208 =

        # UFold: 8,641,377

        self.embed_reduction = embed_reduction

        self.main_dim = 64

        first_layer = nn.Sequential(
            nn.Conv2d(self.embed_reduction*2, self.main_dim, kernel_size=1),
        )

        self.num_res_layers = 32
        res_layers = []
        for i in range(self.num_res_layers):
            dilation = pow(2, (i % 3))
            res_layers.append(MyBasicResBlock(inplanes=self.main_dim, planes=self.main_dim, dilation=dilation))
        res_layers = nn.Sequential(*res_layers)

        final_layer = nn.Conv2d(self.main_dim, num_classes, kernel_size=3, padding=1)    # CJY 2022.12.20  2->1

        layers = OrderedDict()
        layers["first"] = first_layer
        layers["resnet"] = res_layers
        layers["final"] = final_layer

        self.proj = nn.Sequential(layers)


# Res Block
from torch import Tensor
from typing import Type, Any, Callable, Union, List, Optional

def conv3x3(in_planes: int, out_planes: int, stride: int = 1, groups: int = 1, dilation: int = 1) -> nn.Conv2d:
    """3x3 convolution with padding"""
    #return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
    #                 padding=dilation, groups=groups, bias=False, dilation=dilation)
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes: int, out_planes: int, stride: int = 1) -> nn.Conv2d:
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)

class MyBasicResBlock(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None
    ) -> None:
        super(MyBasicResBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        # cjy commented
        #if dilation > 1:
        #    raise NotImplementedError("Dilation > 1 not supported in BasicBlock")

        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.bn1 = norm_layer(inplanes)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.dropout = nn.Dropout(p=0.3)
        self.relu2 = nn.ReLU(inplace=True)
        #self.conv2 = conv3x3(planes, planes)
        self.conv2 = conv3x3(planes, planes, dilation=dilation)

        #self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.bn1(x)
        out = self.relu1(out)
        out = self.conv1(out)
        out = self.dropout(out)
        out = self.relu2(out)
        out = self.conv2(out)
        #out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        #out = self.relu(out)

        return out


class MyBasicResBlock_BN(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None
    ) -> None:
        super(MyBasicResBlock_BN, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        # cjy commented
        #if dilation > 1:
        #    raise NotImplementedError("Dilation > 1 not supported in BasicBlock")

        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.bn1 = norm_layer(inplanes)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv1 = conv3x3(inplanes, planes, stride)
        #self.dropout = nn.Dropout(p=0.3)
        self.bn2 = norm_layer(planes)
        self.relu2 = nn.ReLU(inplace=True)
        #self.conv2 = conv3x3(planes, planes)
        self.conv2 = conv3x3(planes, planes, dilation=dilation)


        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.bn1(x)
        out = self.relu1(out)
        out = self.conv1(out)
        out = self.bn2(out)
        #out = self.dropout(out)
        out = self.relu2(out)
        out = self.conv2(out)


        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        #out = self.relu(out)

        return out

class MyBasicResBlock_KS1(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None
    ) -> None:
        super(MyBasicResBlock_KS1, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        # cjy commented
        #if dilation > 1:
        #    raise NotImplementedError("Dilation > 1 not supported in BasicBlock")

        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.bn1 = norm_layer(inplanes)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv1 = conv1x1(inplanes, planes, stride)
        self.dropout = nn.Dropout(p=0.3)
        self.relu2 = nn.ReLU(inplace=True)
        #self.conv2 = conv3x3(planes, planes)
        self.conv2 = conv1x1(planes, planes,)

        #self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.bn1(x)
        out = self.relu1(out)
        out = self.conv1(out)
        out = self.dropout(out)
        out = self.relu2(out)
        out = self.conv2(out)
        #out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        #out = self.relu(out)

        return out


# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

proteinseq_toks = {
    'toks': ['L', 'A', 'G', 'V', 'S', 'E', 'R', 'T', 'I', 'D', 'P', 'K', 'Q', 'N', 'F', 'Y', 'M', 'H', 'W', 'C', 'X', 'B', 'U', 'Z', 'O', '.', '-']
}

rnaseq_toks = {
    'toks': ['A', 'C', 'G', 'U', 'R', 'Y', 'K', 'M', 'S', 'W', 'B', 'D', 'H', 'V', 'N', '-']
}

rnaseq_3mer_toks = {
    'toks': [
        "GAG", "AAG", "GAA", "CUG", "CAG", "GAU", "AAA", "GUG",
        "GAC", "AUG", "GCC", "AAC", "GCU", "AAU", "AUC", "UUC",
        "GGA", "AUU", "GGC", "UUU", "CCA", "AGC", "GCA", "UCU",
        "CUC", "ACC", "CAA", "CCU", "UCC", "ACA", "UUG", "GUU",
        "CUU", "UAC", "ACU", "CCC", "UCA", "GUC", "GGU", "CAC",
        "AGU", "UAU", "AGA", "CAU", "GGG", "UGG", "UGC", "AGG",
        "UGU", "AUA", "CGC", "UUA", "GCG", "CGG", "CCG", "GUA",
        "CUA", "ACG", "UCG", "CGA", "CGU", "UGA", "UAA", "UAG",
    ]
}


# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from setuptools import setup


with open("fm/version.py") as infile:
    exec(infile.read())

with open("README.md") as f:
    readme = f.read()


setup(
    name="rna-fm",
    version=version,
    description="RNA Foundation Model (rna-fm): Pretrained language models for RNAs. From CUHK AIH Lab.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="CUHK AIH Lab",
    url="https://github.com/ml4bio/RNA-FM",
    license="MIT",
    packages=["fm", "fm/model", "fm/downstream", "fm/downstream/pairwise_predictor"],
    data_files=[("source_docs/fm", ["LICENSE", "README.md"])],
    zip_safe=True,
    install_requires = [
        'numpy',
        'pandas',
        'tqdm',
        'scikit-learn',
        'ptflops',
    ],
)


import torch.nn as nn

def weights_init_kaiming(m):
    classname = m.__class__.__name__
    if classname.find('Linear') != -1:
        nn.init.kaiming_normal_(m.weight, a=0, mode='fan_out')
        nn.init.constant_(m.bias, 0.0)
    elif classname.find('BasicConv') != -1:   # for googlenet
        pass
    elif classname.find('Conv') != -1:
        nn.init.kaiming_normal_(m.weight, a=0, mode='fan_in')
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.0)
    elif classname.find('BatchNorm') != -1:
        if m.affine:
            nn.init.constant_(m.weight, 1.0)
            nn.init.constant_(m.bias, 0.0)

def weights_init_classifier(m):
    classname = m.__class__.__name__
    if classname.find('Linear') != -1:
        nn.init.normal_(m.weight, std=0.001)
        if isinstance(m.bias, nn.Parameter):
            nn.init.constant_(m.bias, 0.0)

def weights_init_toehold_switch(m):
    classname = m.__class__.__name__
    if classname.find('Linear') != -1:
        #nn.init.kaiming_normal_(m.weight, a=0, mode='fan_out')
        #nn.init.constant_(m.bias, 0.0)
        nn.init.xavier_uniform_(m.weight, nn.init.calculate_gain('relu'))
        nn.init.constant_(m.bias, 0.0)
    elif classname.find('BatchNorm') != -1:
        if m.affine:
            nn.init.constant_(m.weight, 1.0)
            nn.init.constant_(m.bias, 0.0)


# encoding: utf-8
"""
@author:  Jiayang Chen
@contact: yjcmydkzgj@gmail.com
"""

# import fm
# def choose_backbone(backbone_name):
#     if backbone_name == 'rna-fm':
#         backbone, backbone_alphabet = fm.pretrained.rna_fm_t12()
#     else:
#         raise Exception("Wrong Backbone Type! {}".format(backbone_name))
#
#     return backbone, backbone_alphabet

from..pretrained import *

def choose_backbone(backbone_name):
    if backbone_name == 'rna-fm':
        backbone, backbone_alphabet = rna_fm_t12()
    else:
        raise Exception("Wrong Backbone Type! {}".format(backbone_name))

    return backbone, backbone_alphabet